import search


def insertion_sort(t):
    comps = 0
    moves = 0
    if len(t) <= 1:
        return t, 0, 0
    for i in range(1, len(t)):
        j = i
        # swap until right position found...
        comps += 1
        while j >= 1 and t[j - 1] > t[j]:
            (t[j - 1], t[j]) = (t[j], t[j - 1])
            moves += 3
            j = j - 1
            comps += 1
    return t, comps, moves


def insertion_sort_optimized(t):
    if len(t) <= 1:
        return t, 0, 0
    comps = 0
    moves = 0
    for i in range(1, len(t)):
        j = i
        # find position
        p, c, m = search.dicho_find_pos(t, t[i], 0, i)
        comps += c
        moves += m
        # shift until position
        moves += 1
        v = t[i]
        for j in range(i, p, -1):
            moves += 1
            t[i] = t[i - 1]
        moves += 1
        t[p] = v
    return t, comps, moves


def selection_sort(t):
    if len(t) <= 1:
        return t, 0, 0
    comps = 0
    moves = 0
    for i in range(1, len(t)):
        min_pos, c, s = search.search_minimum_position_with_bounds(t, i, len(t))
        comps += c
        moves += s
        comps += 1
        if t[i - 1] > t[min_pos]:
            # swap...
            (t[i - 1], t[min_pos]) = (t[min_pos], t[i - 1])
            moves += 3
    return t, comps, moves


def bubble_sort(t):
    if len(t) <= 1:
        return t, 0, 0
    comps = 0
    moves = 0
    swapped = True
    i = 1
    while swapped:
        swapped = False
        for i in range(1, len(t)):
            comps += 1
            if t[i] < t[i - 1]:
                moves += 3
                t[i - 1], t[i] = t[i], t[i - 1]
                swapped = True
    return t, comps, moves


def merge(a, begin, middle, end):
    l = [0] * (middle - begin)
    r = [0] * (end - middle)
    comps = 0
    moves = 0
    for i in range(begin, middle):
        moves += 1
        l[i - begin] = a[i]
    for i in range(middle, end):
        moves += 1
        r[i - middle] = a[i]
    k = begin
    li = 0
    ri = 0
    while li < middle - begin and ri < end - middle:
        comps += 1
        if l[li] < r[ri]:
            moves += 1
            a[k] = l[li]
            li += 1
        else:
            moves += 1
            a[k] = r[ri]
            ri += 1
        k += 1
    while li < middle - begin:
        moves += 1
        a[k] = l[li]
        li += 1
        k += 1
    while ri < end - middle:
        moves += 1
        a[k] = r[ri]
        ri += 1
        k += 1
    return a, comps, moves


def merge_sort_with_bounds(a, begin, end):
    if end - begin <= 1:
        return a, 0, 0
    middle = (end + begin) // 2
    a, cl, ml = merge_sort_with_bounds(a, begin, middle)
    a, cr, mr = merge_sort_with_bounds(a, middle, end)
    a, cm, mm = merge(a, begin, middle, end)
    return a, cl + cr + cm, ml + mr + mm


def merge_sort(a):
    return merge_sort_with_bounds(a, 0, len(a))


def stooge_sort_with_bounds(a, begin, end):
    if end - begin <= 1:
        return a, 0, 0
    comps = 0
    moves = 0
    if end - begin == 2:
        comps += 1
        if a[begin] > a[begin + 1]:
            moves += 3
            a[begin], a[begin + 1] = a[begin + 1], a[begin]
        return a, comps, moves
    third = (end - begin) // 3
    a, c1, m1 = stooge_sort_with_bounds(a, begin, end - third)
    a, c2, m2 = stooge_sort_with_bounds(a, begin + third, end)
    a, c3, m3 = stooge_sort_with_bounds(a, begin, end - third)
    return a, c1 + c2 + c3, m1 + m2 + m3


def stooge_sort(a):
    return stooge_sort_with_bounds(a, 0, len(a))


def quick_choose_pivot(t, begin, end):
    return (begin+end)//2


def quick_sort_partition_in_place(t, begin, end):
    pivot_pos = quick_choose_pivot(t, begin, end)
    moves = 3
    t[begin], t[pivot_pos] = t[pivot_pos],t[begin]
    left = begin+1
    right = end-1
    comps = 0
    while right >= left:
        comps += 1
        if t[left] < t[begin]:
            left += 1
        else:
            comps += 1
            if t[right] >= t[begin]:
                right -= 1
            else:
                moves += 3
                t[left], t[right] = t[right], t[left]
                left += 1
                right -= 1
    moves += 1
    t[begin], t[right] = t[right], t[begin]
    return t, right, comps, moves

# this partition is not in place!
def quick_sort_partition(t, begin, end):
    pivot_pos = quick_choose_pivot(t, begin, end)
    l = []
    r = []
    comps = 0
    moves = 0
    for i in range(begin, pivot_pos):
        comps += 1
        moves += 1
        if t[i] < t[pivot_pos]:
            l.append(t[i])
        else:
            r.append(t[i])
    for i in range(pivot_pos+1, end):
        comps = 0
        moves = 0
        if t[i] < t[pivot_pos]:
            l.append(t[i])
        else:
            r.append(t[i])
    pivot = t[pivot_pos]
    moves += 1
    for i in range(0, len(l)):
        moves += 1
        t[begin+i] = l[i]
    moves += 1
    t[begin+len(l)] = pivot
    for i in range(0, len(r)):
        moves += 1
        t[begin+i+len(l)+1] = r[i]
    return t, begin+len(l), comps, moves


# strategy (recursive)
# - choose an element in the array (called the pivot)
# - arrange elements of the array such that
#   all element smaller than the pivot are on its left
#   and all element larger than the pivot are on its right
#   Note that at this point, pivot will be at its final location
# - recursive call to the two "sub-arrays"
def quick_sort_with_bounds(t, begin, end, partition):
    if end-begin <= 1:
        return t, 0, 0
    t, pivot_pos, c, m = partition(t, begin, end)
    t, cl, ml = quick_sort_with_bounds(t, begin, pivot_pos, partition)
    t, cr, mr = quick_sort_with_bounds(t, pivot_pos+1, end, partition)
    return t, c+cl+cr, m+ml+mr


def quick_sort(t, in_place=True):
    if in_place:
        return quick_sort_with_bounds(t, 0, len(t), quick_sort_partition_in_place)
    else:
        return quick_sort_with_bounds(t, 0, len(t), quick_sort_partition)


def counting_sort(t):
    minimum = min(t)  # linear
    maximum = max(t)  # linear
    counters = [0]*(maximum-minimum+1)
    # count number of times an element is the array
    for e in t:
        counters[e-minimum] += 1
    current_pos = 0
    for i in range(0, len(counters)):
        for j in range(0, counters[i]):
            t[current_pos] = i+minimum
            current_pos += 1
    return t
