import random

import my_random
import sort
import time
import matplotlib.pyplot as plt
import math

MAX_LENGTH = 8000
STEP = 30
SAMPLES = 10


def is_sorted(t):
    if len(t) == 0:
        return True
    for i in range(1, len(t)):
        if t[i - 1] > t[i]:
            return False
    return True


def test_sort_ok(sorting_algo, name):
    print("Testing ", name)
    final_answer = True
    for i in range(0, 100):
        print(".", end="", flush=True)
        a = my_random.random_list(i)
        # print(name, "original",a)
        a, c, s = sorting_algo(a)
        #        print(name, "sorted", a)
        answer = is_sorted(a)
        if not answer:
            print("Ohoh... Problem with", name)
            print("not sorted ", a)
            exit()
        final_answer = final_answer and answer
    print()
    print(final_answer)


def experiences(sorting_algo, size, samples):
    c = 0
    s = 0
    exec_time = 0
    a = [0] * size
    for exp in range(0, samples):
        my_random.random_refill(a)
        start_timestamp = int(time.time() * 1000)
        a, exp_c, exp_s = sorting_algo(a)
        end_timestamp = int(time.time() * 1000)
        exp_time = end_timestamp - start_timestamp
        c += exp_c
        s += exp_s
        exec_time += exp_time
    return exec_time / samples, c / samples, s / samples


def test(sorting_algo, name, max_length, step, samples):
    print("Collecting sorting times for", name)
    lengths = []
    times = []
    comps = []
    moves = []
    for size in range(0, max_length, step):
        exec_time, c, s = experiences(sorting_algo, size, samples)
        # add collected datas...
        lengths.append(size)
        times.append(exec_time)
        comps.append(c)
        moves.append(s)
        print(f"{name} time for array of size {size} is {exec_time}ms.")
    return lengths, times, comps, moves


# try to count-sort an array of birth years...
t = [random.randint(1900, 2022) for _ in range(0, 100)]
print(t)
sort.counting_sort(t)
print(t)

test_sort_ok(sort.insertion_sort_optimized, "Insertion (opt)")
test_sort_ok(sort.selection_sort, "Selection")
test_sort_ok(sort.insertion_sort, "Insertion")
test_sort_ok(sort.merge_sort, "Merge")
test_sort_ok(sort.stooge_sort, "Stooge")
test_sort_ok(sort.quick_sort, "Quick sort")

(ssl, sst, ssc, sss) = test(sort.selection_sort, "Selection sort", min(MAX_LENGTH, 3000), STEP, SAMPLES)
(isl, ist, isc, iss) = test(sort.insertion_sort, "Insertion sort", min(MAX_LENGTH, 3000), STEP, SAMPLES)
(is2l, is2t, is2c, is2s) = test(sort.insertion_sort_optimized, "Insertion sort (opt)", min(MAX_LENGTH, 3000), STEP,
                                SAMPLES)
(msl, mst, msc, msm) = test(sort.merge_sort, "Merge sort", MAX_LENGTH, STEP, SAMPLES)
(qsl, qst, qsc, qsm) = test(sort.quick_sort, "Quick sort (in place)", MAX_LENGTH, STEP, SAMPLES)
# Too long...
# (stsl, stst, stsc, stsm) = test(sort.stooge_sort, "Stooge sort", MAX_LENGTH, STEP, SAMPLES)

fig, ax = plt.subplots(1, 3)

ax[0].set_title("Time")
ax[1].set_title("Comparisons")
ax[2].set_title("Moves")

l1, = ax[0].plot(ssl, sst, color='#ff0000', label="Selection sort")
ax[1].plot(ssl, ssc, color='#ff0000')
ax[2].plot(ssl, sss, color='#ff0000')

l2, = ax[0].plot(isl, ist, color='#00ff00', label="Insertion sort")
ax[1].plot(isl, isc, color='#00ff00')
ax[2].plot(isl, iss, color='#00ff00')

l3, = ax[0].plot(msl, mst, color='#0000ff', label="Merge sort")
ax[1].plot(msl, msc, color='#0000ff')
ax[2].plot(msl, msm, color='#0000ff')

l4, = ax[0].plot(is2l, is2t, color='#ff00ff', label="Insertion sort (opt)")
ax[1].plot(is2l, is2c, color='#ff00ff')
ax[2].plot(is2l, is2s, color='#ff00ff')

my_color = "#00ffff"
l5, = ax[0].plot(qsl, qst, color=my_color, label="Quick sort (in place)")
ax[1].plot(qsl, qsc, color=my_color)
ax[2].plot(qsl, qsm, color=my_color)
# Too long
# l4, = ax[0].plot(msl, mst, color='#00ffff', label="Stooge sort")
# ax[1].plot(msl, msc, color='#00ffff')
# ax[2].plot(msl, msm, color='#00ffff')


# generate x^2
squares = [n * n / 20000 for n in isl]
ax[0].plot(isl, squares, color='#000000', label="n^2")
linear = [n / 10000 for n in qsl]
ax[0].plot(qsl, linear, color='#000000', label="n")
nlogns = [n * math.log(n, 2) / 3000 if n != 0 else 0 for n in qsl]
ax[0].plot(qsl, nlogns, color='#000000', label="n.log(n)")
# squares = [n*n/5 for n in isl]
# ax[0].plot(isl, squares, color='#000000', label="n^2/K'")

plt.legend(handles=[l1, l2, l3, l4, l5])

plt.show()
